
#pragma once
#include <cmath>
#include <cstdlib>
#include <dune/pdelab/common/function.hh>
#include <dune/phasefield/base_parameters.hh>


template<typename GV, typename RF, std::size_t dim_range, typename Params>
class ZeroInitialVector :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim_range>,ZeroInitialVector<GV,RF,dim_range, Params> >
{
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim_range> Traits;

  ZeroInitialVector(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits, ZeroInitialVector>(gv) {}

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y=0;
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class ZeroInitial
  : public ZeroInitialVector<GV,RF,1,Params>
{
public:
  ZeroInitial(const GV & gv, const Params & params) : ZeroInitialVector<GV,RF,1,Params>(gv,params) {}
};

//______________________________________________________________________________________________________________________

#define CONST_INITIAL_U 0

template<typename GV, typename RF, typename Params, const int type>
class ConstInitial :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, ConstInitial<GV,RF,Params, type> >
{
public:
  RF myval;
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  ConstInitial(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,ConstInitial<GV,RF,Params,type> > (gv)
     {
        if(type == CONST_INITIAL_U)
           myval = params.initial.ptree.get("uConst",(RF)1);
        else
           myval = 0;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = myval;
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_RayleighTaylor :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_RayleighTaylor<GV,RF,Params> >
{
  RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_RayleighTaylor(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_RayleighTaylor<GV,RF,Params> > (gv)
     {
       eps = params.eps;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.5*(1-tanh(1/eps/sqrt(2)*(x[1]-0.75-0.01*cos(x[0]*2*3.1415))));
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_3pContactPoint :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_3pContactPoint<GV,RF,Params,Phase> >
{
  RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_3pContactPoint(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_3pContactPoint<GV,RF,Params,Phase> > (gv)
     {
       eps = params.eps;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.0;

    RF d1 = x[0]-0.5;
    RF yx = 0.5*(1+tanh(1/eps/sqrt(2)*d1));

    RF d2 = x[1]-0.75;
    if(Phase == 0)
      d2 = -d2;
    y[0]=0.5*(1+tanh(1/eps/sqrt(2)*d2))*yx;
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_3pFlowTest :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_3pFlowTest<GV,RF,Params,Phase> >
{
  RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_3pFlowTest(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_3pFlowTest<GV,RF,Params,Phase> > (gv)
     {
       eps = params.eps;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.0;

    RF d1 = x[0]-0.3;


    RF d2 = x[1]-0.25;
    RF yy = 0.5*(1+tanh(1/eps/sqrt(2)*d2));
    if(Phase == 0)
      d1 = -d1;
    y=0.5*(1+tanh(1/eps/sqrt(2)*d1))*yy;
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_3pSlipTest :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_3pSlipTest<GV,RF,Params,Phase> >
{
  RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_3pSlipTest(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_3pSlipTest<GV,RF,Params,Phase> > (gv)
     {
       eps = params.eps;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.0;

    RF d2 = x[1]-0.25;

    y = 0.5*(1+tanh(1/eps/sqrt(2)*d2));
  }
};


//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_3pSphereAndDroplet :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_3pSphereAndDroplet<GV,RF,Params,Phase> >
{
  RF eps;
  RF radius;
  RF radius2;
  RF dx;
  RF dy;
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_3pSphereAndDroplet(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_3pSphereAndDroplet<GV,RF,Params,Phase> > (gv)
     {
       eps = params.eps;
       radius = params.initial.ptree.get("Radius",0.1);
       radius2 = params.initial.ptree.get("Radius2",0.15);
       dx = params.initial.ptree.get("dx",0.0);
       dy = params.initial.ptree.get("dx",0.15);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.0;

    RF dist = sqrt( (x[0]-1)*(x[0]-1) + (x[1]-0.5)*(x[1]-0.5) );
    RF dist2 = sqrt( (x[0]-1-dx)*(x[0]-1-dx) + (x[1]-0.5-dy)*(x[1]-0.5-dy) );
    RF y1 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist-radius)));
    RF y2 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist2-radius2)));

    if(Phase == 0)
    {
      y[0] = y1 * y2;
    }
    if(Phase == 1)
    {
      y[0] = y1 * (1-y2);
    }
  }
};
//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_Sphere :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_Sphere<GV,RF,Params> >
{
  RF eps;
  RF radius;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_Sphere(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_Sphere<GV,RF,Params> > (gv)
     {
       eps = params.eps;
       radius = params.initial.ptree.get("Radius",0.2);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist = sqrt( (x[0]-0.5)*(x[0]-0.5) + (x[1]-0.5)*(x[1]-0.5) );
    y[0] = 0.5*(1+tanh(1/eps/sqrt(2)*(dist-radius)));
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_Rectangle :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_Rectangle<GV,RF,Params> >
{
  RF xcenter;
  RF ycenter;
  RF eps;
  RF radius;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_Rectangle(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_Rectangle<GV,RF,Params> > (gv)
     {
       eps = params.eps;
       radius = params.initial.ptree.get("Radius",0.2);
       xcenter = params.initial.ptree.get("xCenter",0.5);
       ycenter = params.initial.ptree.get("yCenter",0.5);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist1 = std::abs(x[0]-xcenter);
    RF dist2 = std::abs(x[1]-ycenter);
    RF m1 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist1-radius)));
    RF m2 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist2-radius)));
    y[0] = 1-(1-m1)*(1-m2);
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_2Precip :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_2Precip<GV,RF,Params> >
{
  RF eps;
  RF x1center;
  RF y1center;
  RF radius1;
  RF x2center;
  RF y2center;
  RF radius2;
  RF x3center;
  RF y3center;
  RF radius3;


public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_2Precip(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_2Precip<GV,RF,Params> > (gv)
     {
       eps = params.eps;
       radius1 = params.initial.ptree.get("Radius1",0.2);
       x1center = params.initial.ptree.get("x1Center",0.5);
       y1center = params.initial.ptree.get("y1Center",0.5);
       radius2 = params.initial.ptree.get("Radius2",0.2);
       x2center = params.initial.ptree.get("x2Center",0.5);
       y2center = params.initial.ptree.get("y2Center",0.5);
       radius3 = params.initial.ptree.get("Radius3",0.2);
       x3center = params.initial.ptree.get("x3Center",0.5);
       y3center = params.initial.ptree.get("y3Center",0.5);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist1 = std::sqrt( (x[0]-x1center)*(x[0]-x1center) + (x[1]-y1center)*(x[1]-y1center) );
    RF y1 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist1-radius1)));

    RF dist2 = std::sqrt( (x[0]-x2center)*(x[0]-x2center) + (x[1]-y2center)*(x[1]-y2center) );
    RF y2 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist2-radius2)));

    RF dist3 = std::sqrt( (x[0]-x3center)*(x[0]-x3center) + (x[1]-y3center)*(x[1]-y3center) );
    RF y3 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist3-radius3)));

    y[0] = y1*y2*y3;
  }
};

template<typename GV, typename RF, typename Params>
class PhiInitial_SummerSchoolFull :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_SummerSchoolFull<GV,RF,Params> >
{

public:
  RF eps;
  RF radius;
  std::vector<RF> xValues;
  std::vector<RF> yValues;


  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_SummerSchoolFull(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_SummerSchoolFull<GV,RF,Params> > (gv)
     {
       eps = params.eps;
       radius = params.initial.ptree.get("Radius",0.2);

       xValues.push_back(0.560); yValues.push_back(0.070);
       xValues.push_back(0.831); yValues.push_back(0.029);
       xValues.push_back(1.353); yValues.push_back(0.008);
       xValues.push_back(1.298); yValues.push_back(0.105);
       xValues.push_back(1.387); yValues.push_back(0.108);
       xValues.push_back(1.959); yValues.push_back(0.014);
       xValues.push_back(2.397); yValues.push_back(-0.001);
       xValues.push_back(3.014); yValues.push_back(0.074);
       xValues.push_back(3.703); yValues.push_back(0.244);
       xValues.push_back(3.808); yValues.push_back(-0.015);
       xValues.push_back(4.190); yValues.push_back(0.071);
       xValues.push_back(4.816); yValues.push_back(0.026);

     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist;
    RF temp;
    RF yc = 1;
    for(int i=0; i<xValues.size(); i++)
    {
      dist = std::sqrt( (x[0]-xValues[i])*(x[0]-xValues[i]) + (x[1]-yValues[i])*(x[1]-yValues[i]) );
      temp = 0.5*(1+tanh(1/eps/sqrt(2)*(dist-radius)));
      yc = yc*temp;
    }
    y[0] = yc;
  }
};

template<typename GV, typename RF, typename Params, const int phase = -1>
class PhiInitial_SummerSchoolPaper :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_SummerSchoolPaper<GV,RF,Params,phase> >
{
public:
  RF eps;
  RF radius;
  std::vector<RF> xValues;
  std::vector<RF> yValues;
private:
  inline static RF PxToCoord_x(const RF px_x)
  {
    return 0.5+(px_x - 150)/((RF)299);
    //return 0.5+(px_x - 150)/((RF)299)-3.5; //-> Christians Vortrag
  }
  inline static RF PxToCoord_y(const RF px_y)
  {
    return (102 - px_y)/((RF)299);
  }

  inline static RF PxToCoord_x2(const RF px_x)
  {
    return (px_x - 163)/((RF)314);
  }
  inline static RF PxToCoord_y2(const RF px_y)
  {
    return (107 - px_y)/((RF)314);
  }

  inline static RF PxToCoord_x3(const RF px_x)
  {
    return (px_x - 391)/((RF)224/1.5);
  }
  inline static RF PxToCoord_y3(const RF px_y)
  {
    return (749 - px_y)/((RF)224/1.5);
  }

  inline static RF PxToCoord_x4(const RF px_x)
  {
    return (px_x - 1081)/((RF)189/0.75);
  }
  inline static RF PxToCoord_y4(const RF px_y)
  {
    return (2037 - px_y)/((RF)189/0.75);
  }
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_SummerSchoolPaper(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_SummerSchoolPaper<GV,RF,Params,phase> > (gv)
     {
       eps = params.eps;
       radius = params.initial.ptree.get("Radius",0.2);



//Experiment 1
       xValues.push_back(PxToCoord_x(114)); yValues.push_back(PxToCoord_y(81)); //Extra Nucleus, to check the influence on Nucleus 1

       xValues.push_back(PxToCoord_x(154)); yValues.push_back(PxToCoord_y(75)); //Nucleus 1
       xValues.push_back(PxToCoord_x(317)); yValues.push_back(PxToCoord_y(85));
       xValues.push_back(PxToCoord_x(364)); yValues.push_back(PxToCoord_y(110));
       xValues.push_back(PxToCoord_x(556)); yValues.push_back(PxToCoord_y(65));

       xValues.push_back(PxToCoord_x(759)); yValues.push_back(PxToCoord_y(85));

       xValues.push_back(PxToCoord_x(832)); yValues.push_back(PxToCoord_y(159));

       xValues.push_back(PxToCoord_x(938)); yValues.push_back(PxToCoord_y(101));
       xValues.push_back(PxToCoord_x(1100)); yValues.push_back(PxToCoord_y(62));
//  xValues.push_back(PxToCoord_x(1240)); yValues.push_back(PxToCoord_y(70));    //Nuclei für Christians Vortrag
       xValues.push_back(PxToCoord_x(1229)); yValues.push_back(PxToCoord_y(20));
       xValues.push_back(PxToCoord_x(1258)); yValues.push_back(PxToCoord_y(29));
       xValues.push_back(PxToCoord_x(1340)); yValues.push_back(PxToCoord_y(57));
       xValues.push_back(PxToCoord_x(1346)); yValues.push_back(PxToCoord_y(77));
       xValues.push_back(PxToCoord_x(1354)); yValues.push_back(PxToCoord_y(103));

       xValues.push_back(PxToCoord_x(1731)); yValues.push_back(PxToCoord_y(115));
       xValues.push_back(PxToCoord_x(1915)); yValues.push_back(PxToCoord_y(77));

       // Addional Nuclei
       //xValues.push_back(PxToCoord_x(668)); yValues.push_back(PxToCoord_y(63));
       //xValues.push_back(PxToCoord_x(775)); yValues.push_back(PxToCoord_y(143));
       //xValues.push_back(PxToCoord_x(874)); yValues.push_back(PxToCoord_y(139));
       //xValues.push_back(PxToCoord_x(815)); yValues.push_back(PxToCoord_y(36));
       //xValues.push_back(PxToCoord_x(848)); yValues.push_back(PxToCoord_y(53));
       //xValues.push_back(PxToCoord_x(1432)); yValues.push_back(PxToCoord_y(101));

/*
//Experiment 7
      xValues.push_back(PxToCoord_x2(423)); yValues.push_back(PxToCoord_y2(129));
      xValues.push_back(PxToCoord_x2(1636)); yValues.push_back(PxToCoord_y2(98));
*/

/*
//Inflow
      xValues.push_back(PxToCoord_x3(402)); yValues.push_back(PxToCoord_y3(642));
      xValues.push_back(PxToCoord_x3(406)); yValues.push_back(PxToCoord_y3(689));
      xValues.push_back(PxToCoord_x3(416)); yValues.push_back(PxToCoord_y3(696));
      xValues.push_back(PxToCoord_x3(425)); yValues.push_back(PxToCoord_y3(709));
      xValues.push_back(PxToCoord_x3(441)); yValues.push_back(PxToCoord_y3(709));
      xValues.push_back(PxToCoord_x3(468)); yValues.push_back(PxToCoord_y3(719));
      xValues.push_back(PxToCoord_x3(435)); yValues.push_back(PxToCoord_y3(681));
      xValues.push_back(PxToCoord_x3(460)); yValues.push_back(PxToCoord_y3(673));
      xValues.push_back(PxToCoord_x3(476)); yValues.push_back(PxToCoord_y3(672));
      xValues.push_back(PxToCoord_x3(497)); yValues.push_back(PxToCoord_y3(693));
      xValues.push_back(PxToCoord_x3(548)); yValues.push_back(PxToCoord_y3(687));
      xValues.push_back(PxToCoord_x3(594)); yValues.push_back(PxToCoord_y3(680));
*/

/*
//Inflow S1
xValues.push_back(PxToCoord_x4(1090)); yValues.push_back(PxToCoord_y4(1858));
xValues.push_back(PxToCoord_x4(1102)); yValues.push_back(PxToCoord_y4(1885));
xValues.push_back(PxToCoord_x4(1126)); yValues.push_back(PxToCoord_y4(1892));
xValues.push_back(PxToCoord_x4(1134)); yValues.push_back(PxToCoord_y4(1905));
xValues.push_back(PxToCoord_x4(1186)); yValues.push_back(PxToCoord_y4(1901));
xValues.push_back(PxToCoord_x4(1248)); yValues.push_back(PxToCoord_y4(1899));
xValues.push_back(PxToCoord_x4(1296)); yValues.push_back(PxToCoord_y4(1908));
xValues.push_back(PxToCoord_x4(1093)); yValues.push_back(PxToCoord_y4(1951));
xValues.push_back(PxToCoord_x4(1122)); yValues.push_back(PxToCoord_y4(1936));
xValues.push_back(PxToCoord_x4(1143)); yValues.push_back(PxToCoord_y4(1942));
xValues.push_back(PxToCoord_x4(1095)); yValues.push_back(PxToCoord_y4(1975));
xValues.push_back(PxToCoord_x4(1141)); yValues.push_back(PxToCoord_y4(1970));
xValues.push_back(PxToCoord_x4(1155)); yValues.push_back(PxToCoord_y4(1985));
xValues.push_back(PxToCoord_x4(1171)); yValues.push_back(PxToCoord_y4(1967));
xValues.push_back(PxToCoord_x4(1230)); yValues.push_back(PxToCoord_y4(1949));
xValues.push_back(PxToCoord_x4(1249)); yValues.push_back(PxToCoord_y4(1957));
xValues.push_back(PxToCoord_x4(1273)); yValues.push_back(PxToCoord_y4(1963));
xValues.push_back(PxToCoord_x4(1122)); yValues.push_back(PxToCoord_y4(2018));
xValues.push_back(PxToCoord_x4(1134)); yValues.push_back(PxToCoord_y4(2010));
xValues.push_back(PxToCoord_x4(1184)); yValues.push_back(PxToCoord_y4(2021));
xValues.push_back(PxToCoord_x4(1199)); yValues.push_back(PxToCoord_y4(1998));
xValues.push_back(PxToCoord_x4(1283)); yValues.push_back(PxToCoord_y4(2029));
*/

     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist;
    RF temp;
    RF yc = 1;
    for(int i=0; i<xValues.size(); i++)
    {
      dist = std::sqrt( (x[0]-xValues[i])*(x[0]-xValues[i]) + (x[1]-yValues[i])*(x[1]-yValues[i]) );
      temp = 0.5*(1+tanh(1/eps/sqrt(2)*(dist-radius)));
      yc = yc*temp;
    }
    if(phase >= 0)
    {
      RF yf = 0.5*(1+tanh(1/eps/sqrt(2)*(x[0]-0.3)));
      if (phase == 0)
      {

        yc = yc*yf;
      }
      else
      {
        yc = yc*(1-yf);
      }
    }
    y[0] = yc;
  }
};


template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_HainesJump :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_HainesJump<GV,RF,Params,Phase> >
{
  RF eps;
  RF period;
  RF amplitude;
  RF height;
  RF offset;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_HainesJump(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_HainesJump<GV,RF,Params,Phase> > (gv)
     {
       eps = params.eps;
       period = params.initial.ptree.get("Period",1);
       amplitude = params.initial.ptree.get("Amplitude",0.3);
       height = params.initial.ptree.get("Height",1);
       offset = params.initial.ptree.get("Offset",0.1);


     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF temp = (-std::cos(x[0]*2*3.1415*period) + 1)*amplitude/2 + offset;
    RF fluid = 0.5*(1+tanh(1/eps/sqrt(2)*(x[1]-temp)));
    fluid = fluid*0.5*(1+tanh(1/eps/sqrt(2)*(height-x[1]-temp)));

    RF d1 = x[0]-0.25;
    if(Phase == 0)
      d1 = -d1;
    y=0.5*(1+tanh(1/eps/sqrt(2)*d1))*fluid;
  }
};

template<typename GV, typename RF, typename Params>
class PhiInitial_SpinodalDecomposition :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_SpinodalDecomposition<GV,RF,Params> >
{
  RF mean;
  RF deviation;
  int seed;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_SpinodalDecomposition(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_SpinodalDecomposition<GV,RF,Params> > (gv)
     {
       mean = params.initial.ptree.get("Mean",0.5);
       deviation = params.initial.ptree.get("Deviation",0.1);
       seed = params.initial.ptree.get("Seed",1);
       std::srand(seed);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y[0] = mean + deviation*(((RF)std::rand())/RAND_MAX);
  }
};

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_3pBubble :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_3pBubble<GV,RF,Params,Phase> >
{
  RF eps;
  RF r;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_3pBubble(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_3pBubble<GV,RF,Params,Phase> > (gv)
     {
       eps = params.eps;
       r = params.initial.ptree.get("Radius",0.2);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist = std::sqrt((x[0]-0.5)*(x[0]-0.5)+(x[1]-0.5)*(x[1]-0.5));
    RF rf = 0.5*(1+tanh(1/eps/sqrt(2)*(dist-r)));

    if(Phase == 0)
    {
      y = 1-rf;
      return;
    }

    RF fluid = 0.5*(1+tanh(1/eps/sqrt(2)*(x[1]-0.5)));
    y=fluid*rf;
  }
};


//______________________________________________________________________________________________________________________

#include <dune/phasefield/boundary_fn.hh>

template<typename GV, typename RF, typename Params, int dim>
class VelocityInitial_Inflow :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>,
  VelocityInitial_Inflow<GV,RF,Params,dim> >
{
  const Params_Time& time;
  RF meanflow;


public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  VelocityInitial_Inflow(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits,VelocityInitial_Inflow<GV,RF,Params,dim> > (gv)
    , time(params.time)
  {
    meanflow = params.initial.ptree.get("MeanInflow",(RF)0.1);
  }

  template <typename T>
  void setTime(T t){
    time = t;
  }

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y = 0.0;
    if(x[0]<GRID_EPS) //Inflow boundary
    {
      RF factor = meanflow * 6/( (GRID_TOP-GRID_BOTTOM)*(GRID_TOP-GRID_BOTTOM) );
      y[0] = std::max((RF)0, factor * (x[1]-GRID_BOTTOM)*(GRID_TOP-x[1]));
    }
  }
};



template<typename GV, typename RF, typename Params, int dim>
class VelocityInitial_Channelflow :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>,
  VelocityInitial_Channelflow<GV,RF,Params,dim> >
{
  const Params_Time& time;
  RF v;


public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  VelocityInitial_Channelflow(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits,VelocityInitial_Channelflow<GV,RF,Params,dim> > (gv)
    , time(params.time)
  {
    v = params.initial.ptree.get("FlowV",(RF)0.1);
  }

  template <typename T>
  void setTime(T t){
    time = t;
  }

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y = 0.0;
    if(x[1]>1-GRID_EPS) //Flow boundary
    {
      y[0] = v;
    }
  }
};

template<typename GV, typename RF, typename Params>
class UInitial_Inflow :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, UInitial_Inflow<GV,RF,Params> >
{
  RF uConst;
  RF uInflow;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  UInitial_Inflow(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,UInitial_Inflow<GV,RF,Params> > (gv)
     {
       uConst = params.initial.ptree.get("uConst",(RF)0.5);
       uInflow = params.initial.ptree.get("uInflow",(RF)0.5);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = uConst;
    if(x[0]<1e-7)
    {
      y=uInflow;
    }
  }
};

template<typename GV, typename RF, typename Params>
class PInitial_Inflow :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PInitial_Inflow<GV,RF,Params> >
{
  RF pConst;
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PInitial_Inflow(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PInitial_Inflow<GV,RF,Params> > (gv)
     {
       pConst = params.initial.ptree.get("pConst",(RF)0.01);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0;
    if(x[0]<1e-7)
    {
      y=pConst;
    }
  }
};
